package com.sun.org.apache.xerces.internal.impl.dv.xs;

import com.sun.org.apache.xerces.internal.xs.datatypes.XSDouble;
import com.sun.org.apache.xerces.internal.impl.dv.InvalidDatatypeValueException;
import com.sun.org.apache.xerces.internal.impl.dv.ValidationContext;

public class DoubleDV extends TypeValidator
{
    public DoubleDV() {
    }
    
    @Override
    public short getAllowedFacets() {
        return 2552;
    }
    
    @Override
    public Object getActualValue(String content, ValidationContext context) throws InvalidDatatypeValueException {
        try {
            return new XDouble(content);
        }
        catch (NumberFormatException ex) {
            throw new InvalidDatatypeValueException("cvc-datatype-valid.1.2.1", new Object[] { content, "double" });
        }
    }
    
    @Override
    public int compare(Object value1, Object value2) {
        return DoubleDV.XDouble.access$000((DoubleDV.XDouble)value1, (DoubleDV.XDouble)value2);
    }
    
    @Override
    public boolean isIdentical(Object value1, Object value2) {
        return value2 instanceof DoubleDV.XDouble && ((DoubleDV.XDouble)value1).isIdentical((DoubleDV.XDouble)value2);
    }
    
    static boolean isPossibleFP(String val) {
        int length = val.length();
        int i = 0;
        while (i < length) {
            char c = val.charAt(i);
            if ((c < '0' || c > '9') && c != '.' && c != '-' && c != '+' && c != 'E' && c != 'e') {
                return false;
            }
            ++i;
        }
        return true;
    }
    
    private static final class XDouble implements XSDouble
    {
        private final double value;
        private String canonical;
        
        public XDouble(String s) throws NumberFormatException {
            if (DoubleDV.isPossibleFP(s)) {
                this.value = Double.parseDouble(s);
            }
            else if (s.equals("INF")) {
                this.value = Double.POSITIVE_INFINITY;
            }
            else if (s.equals("-INF")) {
                this.value = Double.NEGATIVE_INFINITY;
            }
            else if (s.equals("NaN")) {
                this.value = Double.NaN;
            }
            else {
                throw new NumberFormatException(s);
            }
        }
        
        @Override
        public boolean equals(Object val) {
            if (val == this) {
                return true;
            }
            if (!(val instanceof DoubleDV.XDouble)) {
                return false;
            }
            DoubleDV.XDouble oval = (DoubleDV.XDouble)val;
            return this.value == oval.value || (this.value != this.value && oval.value != oval.value);
        }
        
        @Override
        public int hashCode() {
            if (this.value == 0.0) {
                return 0;
            }
            long v = Double.doubleToLongBits(this.value);
            return (int)(v ^ v >>> 32);
        }
        
        public boolean isIdentical(DoubleDV.XDouble val) {
            if (val == this) {
                return true;
            }
            if (this.value == val.value) {
                return this.value != 0.0 || Double.doubleToLongBits(this.value) == Double.doubleToLongBits(val.value);
            }
            return this.value != this.value && val.value != val.value;
        }
        
        private int compareTo(DoubleDV.XDouble val) {
            double oval = val.value;
            if (this.value < oval) {
                return -1;
            }
            if (this.value > oval) {
                return 1;
            }
            if (this.value == oval) {
                return 0;
            }
            if (this.value != this.value) {
                if (oval != oval) {
                    return 0;
                }
                return 2;
            }
            else {
                return 2;
            }
        }
        
        @Override
        public synchronized String toString() {
            if (this.canonical == null) {
                if (this.value == Double.POSITIVE_INFINITY) {
                    this.canonical = "INF";
                }
                else if (this.value == Double.NEGATIVE_INFINITY) {
                    this.canonical = "-INF";
                }
                else if (this.value != this.value) {
                    this.canonical = "NaN";
                }
                else if (this.value == 0.0) {
                    this.canonical = "0.0E1";
                }
                else {
                    this.canonical = Double.toString(this.value);
                    if (this.canonical.indexOf(69) == -1) {
                        int len = this.canonical.length();
                        char[] chars = new char[len + 3];
                        this.canonical.getChars(0, len, chars, 0);
                        int edp = (chars[0] == '-') ? 2 : 1;
                        if (this.value >= 1.0 || this.value <= -1.0) {
                            int i;
                            int dp;
                            for (dp = (i = this.canonical.indexOf(46)); i > edp; --i) {
                                chars[i] = chars[i - 1];
                            }
                            chars[edp] = '.';
                            while (chars[len - 1] == '0') {
                                --len;
                            }
                            if (chars[len - 1] == '.') {
                                ++len;
                            }
                            chars[len++] = 'E';
                            int shift = dp - edp;
                            chars[len++] = (char)(shift + 48);
                        }
                        else {
                            int nzp;
                            for (nzp = edp + 1; chars[nzp] == '0'; ++nzp) {}
                            chars[edp - 1] = chars[nzp];
                            chars[edp] = '.';
                            for (int i = nzp + 1, j = edp + 1; i < len; ++i, ++j) {
                                chars[j] = chars[i];
                            }
                            len -= nzp - edp;
                            if (len == edp + 1) {
                                chars[len++] = '0';
                            }
                            chars[len++] = 'E';
                            chars[len++] = '-';
                            int shift = nzp - edp;
                            chars[len++] = (char)(shift + 48);
                        }
                        this.canonical = new String(chars, 0, len);
                    }
                }
            }
            return this.canonical;
        }
        
        @Override
        public double getValue() {
            return this.value;
        }
        
        static int access$000(DoubleDV.XDouble x0, DoubleDV.XDouble x1) {
            return x0.compareTo(x1);
        }
    }
}
